#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/fb.h>

#include "gp2x.h"
#include "fonts.h"
#include "usbjoy.h"


volatile unsigned short *memregs;


static void bah(int i)
{
	printf("bah: %i, errno: %i: ", i, errno);
	perror("");
	exit(1);
}


static void gp2x_text(unsigned char *screen, int x, int y, const char *text, int color)
{
	int i,l;

	screen = screen + x + y*320;

	for (i = 0; i < strlen(text); i++)
	{
		for (l=0;l<8;l++)
		{
			screen[l*320+0]=(fontdata8x8[((text[i])*8)+l]&0x80)?color:0;
			screen[l*320+1]=(fontdata8x8[((text[i])*8)+l]&0x40)?color:0;
			screen[l*320+2]=(fontdata8x8[((text[i])*8)+l]&0x20)?color:0;
			screen[l*320+3]=(fontdata8x8[((text[i])*8)+l]&0x10)?color:0;
			screen[l*320+4]=(fontdata8x8[((text[i])*8)+l]&0x08)?color:0;
			screen[l*320+5]=(fontdata8x8[((text[i])*8)+l]&0x04)?color:0;
			screen[l*320+6]=(fontdata8x8[((text[i])*8)+l]&0x02)?color:0;
			screen[l*320+7]=(fontdata8x8[((text[i])*8)+l]&0x01)?color:0;
		}
		screen += 8;
	}
}


void gp2x_text_lim(unsigned char *screen, int x, int y, const char *texto, int color, int max)
{
	char    buffer[320/8+1];

	strncpy(buffer, texto, 320/8);
	if (max > 320/8) max = 320/8;
	if (max < 0) max = 0;
	buffer[max] = 0;

	gp2x_text(screen,x,y,buffer,color);
}



unsigned int gp2x_joystick_read(unsigned int *rawval)
{
  	unsigned int value=(memregs[0x1198>>1] & 0x00FF);
	*rawval = ~value & 0xff;
  	if(value==0xFD) value=0xFA;
  	if(value==0xF7) value=0xEB;
  	if(value==0xDF) value=0xAF;
  	if(value==0x7F) value=0xBE;
  	value = ~((memregs[0x1184>>1] & 0xFF00) | value | (memregs[0x1186>>1] << 16));

	return value;
}

#define DOWN(key) (keys & GP2X_##key)
#define DOWN_O(key) (oldkeys & GP2X_##key)


int main()
{
	int memdev, fbdev, ret, i, v, y, joy;
	struct fb_fix_screeninfo fixed_info;
	unsigned char *scr;
	unsigned int keys, rawkeys, oldkeys=0;
	char buff[16];

	fbdev = open("/dev/fb0", O_RDWR);
	if (fbdev == -1) bah(1);
	memdev = open("/dev/mem", O_RDWR);
	if (memdev == -1) bah(2);

	ret = ioctl(fbdev, FBIOGET_FSCREENINFO, &fixed_info);
	if (ret == -1) bah(3);
	scr = mmap(0, 320*240, PROT_READ|PROT_WRITE, MAP_SHARED, memdev, fixed_info.smem_start);
	if (scr == MAP_FAILED) bah(4);
	memregs = mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, memdev, 0xc0000000);
	if (memregs == MAP_FAILED) bah(5);

	gp2x_usbjoy_init();

	memregs[0x28DA>>1]=(1<<9)|0xAB;		/* bpp... */
	memregs[0x290C>>1]=320;			/* line width in bytes */
	memregs[0x290E>>1]=(unsigned short)(fixed_info.smem_start);
	memregs[0x2910>>1]=(unsigned short)(fixed_info.smem_start >> 16);
	memregs[0x2912>>1]=(unsigned short)(fixed_info.smem_start);
	memregs[0x2914>>1]=(unsigned short)(fixed_info.smem_start >> 16);

	/* set palette */
	memregs[0x2958>>1]=0;
	memregs[0x295A>>1]=0x0000;
	memregs[0x295A>>1]=0x0000;
	memregs[0x295A>>1]=0xffff;
	memregs[0x295A>>1]=0x00ff;
	memregs[0x295A>>1]=0x0000;
	memregs[0x295A>>1]=0x00ff;

	memset(scr, 0, 320*240);

	while (~memregs[0x1184>>1] & 0x7f00);

	while (1)
	{
		/* vsync.. */
		v = memregs[0x1182>>1];
		while (!((v ^ memregs[0x1182>>1]) & 0x10));

		keys = gp2x_joystick_read(&rawkeys);
		if (DOWN(L) && DOWN(R) && DOWN(START)) break;
		if (DOWN(L) && DOWN(R) && DOWN(SELECT) &&
				!(DOWN_O(L) && DOWN_O(R) && DOWN_O(SELECT)))
		{
			/* rescan */
			gp2x_usbjoy_deinit();
			gp2x_usbjoy_init();
			memset(scr, 0, 320*240);
		}

		buff[1] = 0;
		buff[0] = '0' + num_of_joys;
		gp2x_text(scr,   0, 0, "L+R+START=exit  L+R+SELECT=rescan joys", 1);
		gp2x_text(scr,   0, 8, "joys:", 2);
		gp2x_text(scr, 6*8, 8, buff, 2);

		/* GP2x joy */
		gp2x_text(scr, 10*8, 16, "A", DOWN(UP)    ? 2 : 1);
		gp2x_text(scr, 31*8, 16, "Y", DOWN(Y)     ? 2 : 1);
		gp2x_text(scr,  8*8, 24, "<", DOWN(LEFT)  ? 2 : 1);
		gp2x_text(scr, 10*8, 24, "C", DOWN(PUSH)  ? 2 : 1);
		gp2x_text(scr, 12*8, 24, ">", DOWN(RIGHT) ? 2 : 1);
		gp2x_text(scr, 29*8, 24, "A", DOWN(A)     ? 2 : 1);
		gp2x_text(scr, 33*8, 24, "B", DOWN(B)     ? 2 : 1);
		gp2x_text(scr, 10*8, 32, "V", DOWN(DOWN)  ? 2 : 1);
		gp2x_text(scr, 31*8, 32, "X", DOWN(X)     ? 2 : 1);
		gp2x_text(scr,  6*8, 16, "L", DOWN(L)     ? 2 : 1);
		gp2x_text(scr, 35*8, 16, "R", DOWN(R)     ? 2 : 1);
		gp2x_text(scr,  6*8, 40, "VOL-",   DOWN(VOL_DOWN) ? 2 : 1);
		gp2x_text(scr, 11*8, 40, "VOL+",   DOWN(VOL_UP)   ? 2 : 1);
		gp2x_text(scr, 25*8, 40, "SELECT", DOWN(SELECT)   ? 2 : 1);
		gp2x_text(scr, 32*8, 40, "START",  DOWN(START)    ? 2 : 1);
		gp2x_text(scr,  9*8, 18, ".", (rawkeys & 0x02) ? 2 : 0);
		gp2x_text(scr, 11*8, 18, ".", (rawkeys & 0x80) ? 2 : 0);
		gp2x_text(scr,  9*8, 28, ".", (rawkeys & 0x08) ? 2 : 0);
		gp2x_text(scr, 11*8, 28, ".", (rawkeys & 0x20) ? 2 : 0);

		oldkeys = keys;

		/* usbjoys.. */
		y = 7*8;
		gp2x_usbjoy_update();
		for (joy = 0; joy < num_of_joys && joy < 4; joy++)
		{
			gp2x_text_lim(scr, 0, y, joy_name(joys[joy]), 1, 40);
			y += 8;
			keys = gp2x_usbjoy_check2(joy);

			gp2x_text(scr, 10*8, y+16, "A", joy_getaxe(JOYUP, joys[joy])    ? 2 : 1);
			gp2x_text(scr,  8*8, y+24, "<", joy_getaxe(JOYLEFT, joys[joy])  ? 2 : 1);
			gp2x_text(scr, 12*8, y+24, ">", joy_getaxe(JOYRIGHT, joys[joy]) ? 2 : 1);
			gp2x_text(scr, 10*8, y+32, "V", joy_getaxe(JOYDOWN, joys[joy])  ? 2 : 1);

			for (i = 0; i < joy_buttons(joys[joy]) && i < 32; i++)
			{
				sprintf(buff, "%02i", i+1);
				gp2x_text(scr, (i&7)*3*8 + 14*8, y + (i&~7), buff, (keys & (1 << (i+4))) ? 2 : 1);
			}

			sprintf(buff, "%6i", joys[joy]->axevals[0]);
			gp2x_text(scr, 0, (y += 8), buff, 1);
			sprintf(buff, "%6i", joys[joy]->axevals[1]);
			gp2x_text(scr, 0, (y += 8), buff, 1);
			sprintf(buff, "btns: %i", joy_buttons(joys[joy]));
			gp2x_text(scr, 0, (y += 8), buff, 1);
			sprintf(buff, "axes: %i", joy_axes(joys[joy]));
			gp2x_text(scr, 0, (y += 8), buff, 1);
			y += 2*8;
		}
	}


	gp2x_usbjoy_deinit();

	munmap(scr, 320*240);
	munmap((void *) memregs, 0x10000);
	close(fbdev);
	close(memdev);

	chdir("/usr/gp2x");
	execl("gp2xmenu", "gp2xmenu", NULL);

	return 0;
}

